#=============================================================================
# Copyright (c) 2018-2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==========================================================================

#[=======================================================================[

nvCOMP is a CUDA library that features generic compression interfaces to enable
developers to use high-performance GPU compressors and decompressors in their applications.

Result Variables
^^^^^^^^^^^^^^^^

This module will construct the following targets::
  nvcomp::nvcomp

This module will set the following variables::

  NVCOMP_FOUND

#]=======================================================================]


####### Expanded from @PACKAGE_INIT@ by configure_package_config_file() #######
####### Any changes to this file will be overwritten by the next CMake run ####
####### The input file was nvcomp-config.cmake.in                            ########

get_filename_component(PACKAGE_PREFIX_DIR "${CMAKE_CURRENT_LIST_DIR}/../../../../" ABSOLUTE)

macro(set_and_check _var _file)
  set(${_var} "${_file}")
  if(NOT EXISTS "${_file}")
    message(FATAL_ERROR "File or directory ${_file} referenced by variable ${_var} does not exist !")
  endif()
endmacro()

macro(check_required_components _NAME)
  foreach(comp ${${_NAME}_FIND_COMPONENTS})
    if(NOT ${_NAME}_${comp}_FOUND)
      if(${_NAME}_FIND_REQUIRED_${comp})
        set(${_NAME}_FOUND FALSE)
      endif()
    endif()
  endforeach()
endmacro()

####################################################################################

set(nvcomp_VERSION 4.2.0.11)

## Find header and lib directories via known header and lib names
## Note: we are resorting to the device lib as that is always shipped
##       in the same form
set(_nvcomp_search_header_name nvcomp.h)
if(MSVC)
    set(_nvcomp_search_lib_name nvcomp_device_static.lib)
    set(_nvcomp_search_bin_name nvcomp.dll)
else()
    set(_nvcomp_search_lib_name libnvcomp_device_static.a)
endif()

file(REAL_PATH "${CMAKE_CURRENT_LIST_DIR}" _nvcomp_cmake_config_realpath) # path to the location of nvcomp-config.cmake, with symlinks resolved
file(REAL_PATH "../../" _nvcomp_search_prefix BASE_DIRECTORY "${_nvcomp_cmake_config_realpath}")

if(MSVC)
    find_path(nvcomp_INCLUDE_DIR
              NAMES ${_nvcomp_search_header_name}
              PATHS ${_nvcomp_search_prefix}
              PATH_SUFFIXES
                ../include
                ../../include
              NO_DEFAULT_PATH
              NO_CACHE)

    find_path(nvcomp_LIBRARY_DIR
              NAMES ${_nvcomp_search_lib_name}
              PATHS ${_nvcomp_search_prefix}
              PATH_SUFFIXES
                ../lib
                ../lib/12
                ../../lib
                ../../lib/12
              NO_DEFAULT_PATH
              NO_CACHE)

    find_path(nvcomp_BINARY_DIR
              NAMES ${_nvcomp_search_bin_name}
              PATHS ${_nvcomp_search_prefix}
              PATH_SUFFIXES
                ../bin
                ../bin/12
                ../../bin
                ../../bin/12
              NO_DEFAULT_PATH
              NO_CACHE)
else()
    find_path(nvcomp_INCLUDE_DIR
              NAMES ${_nvcomp_search_header_name}
              PATHS ${_nvcomp_search_prefix}
              PATH_SUFFIXES
                ../include
                ../include/nvcomp # Symlink
                ../../include
                ../../include/nvcomp # Symlink
                ../../../include
                ../../../include/nvcomp # Symlink
                ../../../../include # Debian/Ubuntu <multiarch-tuple>
                ../../../../include/nvcomp # Debian/Ubuntu <multiarch-tuple> + Symlink
              NO_DEFAULT_PATH
              NO_CACHE)

    find_path(nvcomp_LIBRARY_DIR
              NAMES ${_nvcomp_search_lib_name}
              PATHS ${_nvcomp_search_prefix}
              PATH_SUFFIXES
                ./
                ../lib
                ../lib/12
                ../lib64
                ../lib64/12
              NO_DEFAULT_PATH
              NO_CACHE)

endif()

# Check headers and library directories are found
if(NOT EXISTS "${nvcomp_INCLUDE_DIR}")
    message(FATAL_ERROR "Header directory containing file ${_nvcomp_search_header_name} was not found relative to ${CMAKE_CURRENT_LIST_DIR}!")
endif()
if(NOT EXISTS "${nvcomp_LIBRARY_DIR}")
    message(FATAL_ERROR "Library directory containing file ${_nvcomp_search_lib_name} was not found relative to ${CMAKE_CURRENT_LIST_DIR}!")
endif()

# Normalize nvcomp_INCLUDE_DIR and set cache variable
get_filename_component(nvcomp_INCLUDE_DIR "${nvcomp_INCLUDE_DIR}" ABSOLUTE)
set(nvcomp_INCLUDE_DIR ${nvcomp_INCLUDE_DIR} CACHE STRING "Location of nvcomp headers")

# Normalize nvcomp_LIBRARY_DIR and set cache variable
get_filename_component(nvcomp_LIBRARY_DIR "${nvcomp_LIBRARY_DIR}" ABSOLUTE)
set(nvcomp_LIBRARY_DIR ${nvcomp_LIBRARY_DIR} CACHE STRING "Location of nvcomp libraries")

if (EXISTS "${nvcomp_BINARY_DIR}")
  # Normalize nvcomp_BINARY_DIR and set cache variable
  get_filename_component(nvcomp_BINARY_DIR "${nvcomp_BINARY_DIR}" ABSOLUTE)
  set(nvcomp_BINARY_DIR ${nvcomp_BINARY_DIR} CACHE STRING "Location of nvcomp binaries")
endif()

unset(_nvcomp_search_header_name)
unset(_nvcomp_search_lib_name)
if (MSVC)
    unset(_nvcomp_search_bin_name)
endif()
unset(_nvcomp_search_prefix)

find_package(CUDAToolkit REQUIRED)

include("${CMAKE_CURRENT_LIST_DIR}/nvcomp-targets-common.cmake")
## PackageConfig COMPONENTS
# Find installed components
set(_components)
set(nvcomp_device_static_FOUND 1)
list(APPEND _components nvcomp_device_static)

if (EXISTS "${CMAKE_CURRENT_LIST_DIR}/nvcomp-targets-static.cmake")
	include("${CMAKE_CURRENT_LIST_DIR}/nvcomp-targets-static.cmake"
	         OPTIONAL RESULT_VARIABLE _static_targets_found)
	set(nvcomp_static_FOUND 1)
	set(nvcomp_cpu_static_FOUND 1)
	list(APPEND _components nvcomp_static)
	list(APPEND _components nvcomp_cpu_static)
endif()
if (EXISTS "${CMAKE_CURRENT_LIST_DIR}/nvcomp-targets-dynamic.cmake")
	include("${CMAKE_CURRENT_LIST_DIR}/nvcomp-targets-dynamic.cmake")
	set(nvcomp_FOUND 1)
	set(nvcomp_cpu_FOUND 1)
	list(APPEND _components nvcomp)
	list(APPEND _components nvcomp_cpu)
endif()


check_required_components(nvcomp)

## Report status
if(NOT ${CMAKE_FIND_PACKAGE_NAME}_FIND_QUIETLY)
    if (EXISTS "${nvcomp_BINARY_DIR}")
        message(STATUS
"Found nvcomp: (Version:${nvcomp_VERSION}
               CMakePackageDir:${CMAKE_CURRENT_LIST_DIR}
               IncludeDir:${nvcomp_INCLUDE_DIR}
               LibraryDir:${nvcomp_LIBRARY_DIR}
               BinaryDir:${nvcomp_BINARY_DIR}
               ComponentsFound:[${_components}])"
    )
    else()
        message(STATUS
"Found nvcomp: (Version:${nvcomp_VERSION}
               CMakePackageDir:${CMAKE_CURRENT_LIST_DIR}
               IncludeDir:${nvcomp_INCLUDE_DIR}
               LibraryDir:${nvcomp_LIBRARY_DIR}
               ComponentsFound:[${_components}])"
        )
    endif()
endif()
unset(_components)

# The full path to the configuration file is stored in the cmake variable
# <PackageName>_CONFIG.
set(${CMAKE_FIND_PACKAGE_NAME}_CONFIG "${CMAKE_CURRENT_LIST_FILE}")

#include(FindPackageHandleStandardArgs)
#find_package_handle_standard_args(${CMAKE_FIND_PACKAGE_NAME} CONFIG_MODE)
